@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 #
 # Copyright (C) 2008-2009 Edgewall Software
 # All rights reserved.
@@ -13,15 +12,13 @@
 
 """Support classes for generating code from abstract syntax trees."""
 
-try:
-    import ast
-except ImportError:
-    from chameleon import ast25 as ast
-
-import sys
+import ast
+import collections
 import logging
 import weakref
-import collections
+
+
+AST_NONE = ast.Name(id='None', ctx=ast.Load())
 
 node_annotations = weakref.WeakKeyDictionary()
 
@@ -29,9 +26,9 @@ try:
     node_annotations[ast.Name()] = None
 except TypeError:
     logging.debug(
-        "Unable to create weak references to AST nodes. " \
+        "Unable to create weak references to AST nodes. "
         "A lock will be used around compilation loop."
-        )
+    )
 
     node_annotations = {}
 
@@ -69,13 +66,13 @@ def subscript(name, value, ctx):
         value=value,
         slice=ast.Index(value=ast.Str(s=name)),
         ctx=ctx,
-        )
+    )
 
 
 def walk_names(target, mode):
     for node in ast.walk(target):
         if isinstance(node, ast.Name) and \
-               isinstance(node.ctx, mode):
+                isinstance(node.ctx, mode):
             yield node.id
 
 
@@ -123,11 +120,12 @@ def copy(source, target):
     target.__dict__ = source.__dict__
 
 
-def swap(root, replacement, name):
+def swap(body, replacement, name):
+    root = ast.Expression(body=body)
     for node in ast.walk(root):
         if (isinstance(node, ast.Name) and
             isinstance(node.ctx, ast.Load) and
-            node.id == name):
+                node.id == name):
             assert hasattr(replacement, '_fields')
             node_annotations.setdefault(node, replacement)
 
@@ -136,7 +134,7 @@ def marker(name):
     return ast.Str(s="__%s" % name)
 
 
-class Node(object):
+class Node:
     """AST baseclass that gives us a convenient initialization
     method. We explicitly declare and use the ``_fields`` attribute."""
 
@@ -152,12 +150,12 @@ class Node(object):
         """Poor man's single-line pretty printer."""
 
         name = type(self).__name__
-        return '<%s%s at %x>' % (
+        return '<{}{} at {:x}>'.format(
             name,
-            "".join(" %s=%r" % (name, getattr(self, name, "\"?\""))
-                        for name in self._fields),
+            "".join(" {}={!r}".format(name, getattr(self, name, "\"?\""))
+                    for name in self._fields),
             id(self)
-            )
+        )
 
     def extract(self, condition):
         result = []
@@ -208,7 +206,7 @@ class TokenRef(Node):
     _fields = "pos", "length"
 
 
-class ASTCodeGenerator(object):
+class ASTCodeGenerator:
     """General purpose base class for AST transformations.
 
     Every visitor method can be overridden to return an AST node that has been
@@ -235,7 +233,7 @@ class ASTCodeGenerator(object):
         self.code = "\n".join(
             line.strip() and line or ""
             for line in self.lines
-            )
+        )
 
     def _change_indent(self, delta):
         self.indent += delta
@@ -272,7 +270,7 @@ class ASTCodeGenerator(object):
     def visit(self, node):
         if node is None:
             return None
-        if type(node) is tuple:
+        if isinstance(node, tuple):
             return tuple([self.visit(n) for n in node])
         try:
             self.blame_stack.append((node.lineno, node.col_offset,))
@@ -281,7 +279,7 @@ class ASTCodeGenerator(object):
             info = False
         visitor = getattr(self, 'visit_%s' % node.__class__.__name__, None)
         if visitor is None:
-            raise Exception('No handler for ``%s`` (%s).' % (
+            raise Exception('No handler for ``{}`` ({}).'.format(
                 node.__class__.__name__, repr(node)))
         ret = visitor(node)
         if info:
@@ -584,11 +582,7 @@ class ASTCodeGenerator(object):
             self._write(' ')
             self.visit(node.type)
         if getattr(node, 'name', None):
-            if sys.version_info[0] == 2:
-                assert getattr(node, 'type', None)
-                self._write(', ')
-            else:
-                self._write(' as ')
+            self._write(' as ')
             self.visit(node.name)
         self._write(':')
         self._change_indent(1)
@@ -694,7 +688,7 @@ class ASTCodeGenerator(object):
         self._new_line()
         self._write('continue')
 
-    ### EXPRESSIONS
+    # EXPRESSIONS
     def with_parens(f):
         def _f(self, node):
             self._write('(')
@@ -784,6 +778,23 @@ class ASTCodeGenerator(object):
         self.visit(last)
         self._write('}')
 
+    # DictComp(expr key, expr value, comprehension* generators)
+    def visit_DictComp(self, node):
+        self._write('{')
+        self.visit(node.key)
+        self._write(': ')
+        self.visit(node.value)
+        for generator in node.generators:
+            # comprehension = (expr target, expr iter, expr* ifs)
+            self._write(' for ')
+            self.visit(generator.target)
+            self._write(' in ')
+            self.visit(generator.iter)
+            for ifexpr in generator.ifs:
+                self._write(' if ')
+                self.visit(ifexpr)
+        self._write('}')
+
     # ListComp(expr elt, comprehension* generators)
     def visit_ListComp(self, node):
         self._write('[')
@@ -814,6 +825,21 @@ class ASTCodeGenerator(object):
                 self.visit(ifexpr)
         self._write(')')
 
+    # SetComp(expr elt, comprehension* generators)
+    def visit_SetComp(self, node):
+        self._write('{')
+        self.visit(node.elt)
+        for generator in node.generators:
+            # comprehension = (expr target, expr iter, expr* ifs)
+            self._write(' for ')
+            self.visit(generator.target)
+            self._write(' in ')
+            self.visit(generator.iter)
+            for ifexpr in generator.ifs:
+                self._write(' if ')
+                self.visit(ifexpr)
+                self._write('}')
+
     # Yield(expr? value)
     def visit_Yield(self, node):
         self._write('yield')
@@ -865,21 +891,7 @@ class ASTCodeGenerator(object):
             else:
                 self._write('**')
             self.visit(keyword.value)
-        # Attribute removed in Python 3.5
-        if getattr(node, 'starargs', None):
-            if not first:
-                self._write(', ')
-            first = False
-            self._write('*')
-            self.visit(node.starargs)
 
-        # Attribute removed in Python 3.5
-        if getattr(node, 'kwargs', None):
-            if not first:
-                self._write(', ')
-            first = False
-            self._write('**')
-            self.visit(node.kwargs)
         self._write(')')
 
     # Repr(expr value)
@@ -924,20 +936,31 @@ class ASTCodeGenerator(object):
                 for dim in node.slice.elts[1:]:
                     self._write(', ')
                     self.visit(dim)
+        elif isinstance(node.slice, ast.Slice):
+            self.visit_Slice(node.slice, True)
         else:
             self.visit(node.slice)
         self._write(']')
 
     # Slice(expr? lower, expr? upper, expr? step)
-    def visit_Slice(self, node):
-        if getattr(node, 'lower', None) is not None:
-            self.visit(node.lower)
-        self._write(':')
-        if getattr(node, 'upper', None) is not None:
-            self.visit(node.upper)
-        if getattr(node, 'step', None) is not None:
+    def visit_Slice(self, node, subscription=False):
+        if subscription:
+            if getattr(node, 'lower', None) is not None:
+                self.visit(node.lower)
             self._write(':')
-            self.visit(node.step)
+            if getattr(node, 'upper', None) is not None:
+                self.visit(node.upper)
+            if getattr(node, 'step', None) is not None:
+                self._write(':')
+                self.visit(node.step)
+        else:
+            self._write('slice(')
+            self.visit(getattr(node, "lower", None) or AST_NONE)
+            self._write(', ')
+            self.visit(getattr(node, "upper", None) or AST_NONE)
+            self._write(', ')
+            self.visit(getattr(node, "step", None) or AST_NONE)
+            self._write(')')
 
     # Index(expr value)
     def visit_Index(self, node):
@@ -982,6 +1005,7 @@ class ASTCodeGenerator(object):
     def visit_NameConstant(self, node):
         self._write(str(node.value))
 
+
 class AnnotationAwareVisitor(ast.NodeVisitor):
     def visit(self, node):
         annotation = node_annotations.get(node)
@@ -989,7 +1013,7 @@ class AnnotationAwareVisitor(ast.NodeVisitor):
             assert hasattr(annotation, '_fields')
             node = annotation
 
-        super(AnnotationAwareVisitor, self).visit(node)
+        super().visit(node)
 
     def apply_transform(self, node):
         if node not in node_annotations:
