@@ -1,18 +1,25 @@
-import os
-import imp
-import sys
-import py_compile
-import logging
 import functools
+import logging
+import os
+import py_compile
+import shutil
+import sys
 import tempfile
+import warnings
+from importlib.machinery import SourceFileLoader
+from threading import RLock
+
 import pkg_resources
 
-log = logging.getLogger('chameleon.loader')
+from .utils import encode_string
 
-try:
-    str = unicode
-except NameError:
-    basestring = str
+
+lock = RLock()
+acquire_lock = lock.acquire
+release_lock = lock.release
+del lock
+
+log = logging.getLogger('chameleon.loader')
 
 
 def cache(func):
@@ -29,7 +36,14 @@ def abspath_from_asset_spec(spec):
     return pkg_resources.resource_filename(pname, filename)
 
 
-class TemplateLoader(object):
+if os.name == "nt":
+    def abspath_from_asset_spec(spec, f=abspath_from_asset_spec):
+        if spec[1] == ":":
+            return spec
+        return f(spec)
+
+
+class TemplateLoader:
     """Template loader class.
 
     To load templates using relative filenames, pass a sequence of
@@ -48,7 +62,7 @@ class TemplateLoader(object):
     def __init__(self, search_path=None, default_extension=None, **kwargs):
         if search_path is None:
             search_path = []
-        if isinstance(search_path, basestring):
+        if isinstance(search_path, str):
             search_path = [search_path]
         if default_extension is not None:
             self.default_extension = ".%s" % default_extension.lstrip('.')
@@ -69,21 +83,22 @@ class TemplateLoader(object):
         if ':' in spec:
             spec = abspath_from_asset_spec(spec)
 
-        if os.path.isabs(spec):
-            return cls(spec, **self.kwargs)
+        if not os.path.isabs(spec):
+            for path in self.search_path:
+                path = os.path.join(path, spec)
+                if os.path.exists(path):
+                    spec = path
+                    break
+            else:
+                raise ValueError("Template not found: %s." % spec)
 
-        for path in self.search_path:
-            path = os.path.join(path, spec)
-            if os.path.exists(path):
-                return cls(path, **self.kwargs)
-
-        raise ValueError("Template not found: %s." % spec)
+        return cls(spec, search_path=self.search_path, **self.kwargs)
 
     def bind(self, cls):
         return functools.partial(self.load, cls=cls)
 
 
-class MemoryLoader(object):
+class MemoryLoader:
     def build(self, source, filename):
         code = compile(source, filename, 'exec')
         env = {}
@@ -94,9 +109,20 @@ class MemoryLoader(object):
         return None
 
 
-class ModuleLoader(object):
-    def __init__(self, path):
+class ModuleLoader:
+    def __init__(self, path, remove=False):
         self.path = path
+        self.remove = remove
+
+    def __del__(self, shutil=shutil):
+        if not self.remove:
+            return
+        try:
+            shutil.rmtree(self.path)
+        except BaseException:
+            warnings.warn(
+                "Could not clean up temporary file path: %s" %
+                (self.path,))
 
     def get(self, filename):
         path = os.path.join(self.path, filename)
@@ -108,7 +134,7 @@ class ModuleLoader(object):
             log.debug('cache miss: %s' % filename)
 
     def build(self, source, filename):
-        imp.acquire_lock()
+        acquire_lock()
         try:
             d = self.get(filename)
             if d is not None:
@@ -118,16 +144,19 @@ class ModuleLoader(object):
             name = os.path.join(self.path, base + ".py")
 
             log.debug("writing source to disk (%d bytes)." % len(source))
-            fd, fn = tempfile.mkstemp(prefix=base, suffix='.tmp', dir=self.path)
-            temp = os.fdopen(fd, 'w')
+            fd, fn = tempfile.mkstemp(
+                prefix=base, suffix='.tmp', dir=self.path)
+            temp = os.fdopen(fd, 'wb')
+            encoded = source.encode('utf-8')
+            header = encode_string("# -*- coding: utf-8 -*-" + "\n")
 
             try:
                 try:
-                    temp.write("%s\n" % '# -*- coding: utf-8 -*-')
-                    temp.write(source)
+                    temp.write(header)
+                    temp.write(encoded)
                 finally:
                     temp.close()
-            except:
+            except BaseException:
                 os.remove(fn)
                 raise
 
@@ -137,20 +166,15 @@ class ModuleLoader(object):
 
             return self._load(base, name)
         finally:
-            imp.release_lock()
+            release_lock()
 
     def _load(self, base, filename):
-        imp.acquire_lock()
+        acquire_lock()
         try:
             module = sys.modules.get(base)
             if module is None:
-                f = open(filename, 'rb')
-                try:
-                    assert base not in sys.modules
-                    module = imp.load_source(base, filename, f)
-                finally:
-                    f.close()
+                module = SourceFileLoader(base, filename).load_module()
         finally:
-            imp.release_lock()
+            release_lock()
 
         return module.__dict__
