@@ -1,37 +1,33 @@
+import ast
 import re
 import sys
+from ast import Try
 
+from .astutil import Builtin
+from .astutil import ItemLookupOnAttributeErrorVisitor
+from .astutil import Symbol
+from .astutil import load
 from .astutil import parse
 from .astutil import store
-from .astutil import load
-from .astutil import ItemLookupOnAttributeErrorVisitor
 from .codegen import TemplateCodeGenerator
-from .codegen import template
 from .codegen import reverse_builtin_map
-from .astutil import Builtin
-from .astutil import Symbol
-from .exc import ExpressionError
-from .utils import resolve_dotted
-from .utils import Markup
-from .utils import ast
-from .tokenize import Token
-from .parser import substitute
+from .codegen import template
 from .compiler import Interpolator
+from .exc import ExpressionError
+from .parser import substitute
+from .tokenize import Token
+from .utils import ImportableMarker
+from .utils import Markup
+from .utils import lookup_attr
+from .utils import resolve_dotted
 
-try:
-    from .py26 import lookup_attr
-except SyntaxError:
-    from .py25 import lookup_attr
 
+DEFAULT_MARKER = ImportableMarker(__name__, "DEFAULT")
 
 split_parts = re.compile(r'(?<!\\)\|')
-match_prefix = re.compile(r'^\s*([a-z\-_]+):').match
+match_prefix = re.compile(r'^\s*([a-z][a-z0-9\-_]*):').match
 re_continuation = re.compile(r'\\\s*$', re.MULTILINE)
-
-try:
-    from __builtin__ import basestring
-except ImportError:
-    basestring = str
+exc_clear = getattr(sys, "exc_clear", None)
 
 
 def resolve_global(value):
@@ -50,12 +46,14 @@ def test(expression, engine=None, **env):
     module = ast.Module(body)
     module = ast.fix_missing_locations(module)
     env['rcontext'] = {}
+    if exc_clear is not None:
+        env['__exc_clear'] = exc_clear
     source = TemplateCodeGenerator(module).code
     code = compile(source, '<string>', 'exec')
     exec(code, env)
     result = env["result"]
 
-    if isinstance(result, basestring):
+    if isinstance(result, str):
         result = str(result)
 
     return result
@@ -68,10 +66,10 @@ def transform_attribute(node):
         object=node.value,
         name=ast.Str(s=node.attr),
         mode="eval"
-        )
+    )
 
 
-class TalesExpr(object):
+class TalesExpr:
     """Base class.
 
     This class helps implementations for the Template Attribute
@@ -107,11 +105,13 @@ class TalesExpr(object):
     42
     """
 
-    exceptions = NameError, \
-                 ValueError, \
-                 AttributeError, \
-                 LookupError, \
-                 TypeError
+    exceptions = AttributeError, \
+        NameError, \
+        LookupError, \
+        TypeError, \
+        ValueError
+
+    ignore_prefix = True
 
     def __init__(self, expression):
         self.expression = expression
@@ -121,7 +121,7 @@ class TalesExpr(object):
         assignments = []
 
         while remaining:
-            if match_prefix(remaining) is not None:
+            if self.ignore_prefix and match_prefix(remaining) is not None:
                 compiler = engine.parse(remaining)
                 assignment = compiler.assign_value(target)
                 remaining = ""
@@ -135,31 +135,49 @@ class TalesExpr(object):
                     remaining = ""
 
                 expression = expression.replace('\\|', '|')
-                assignment = self.translate(expression, target)
+                assignment = self.translate_proxy(engine, expression, target)
             assignments.append(assignment)
 
         if not assignments:
+            if not remaining:
+                raise ExpressionError("No input:", remaining)
+
             assignments.append(
-                self.translate(remaining, target)
-                )
+                self.translate_proxy(engine, remaining, target)
+            )
 
         for i, assignment in enumerate(reversed(assignments)):
             if i == 0:
                 body = assignment
             else:
-                body = [ast.TryExcept(
+                body = [Try(
                     body=assignment,
                     handlers=[ast.ExceptHandler(
                         type=ast.Tuple(
                             elts=map(resolve_global, self.exceptions),
                             ctx=ast.Load()),
                         name=None,
-                        body=body,
-                        )],
-                    )]
+                        body=body if exc_clear is None else body + [
+                            ast.Expr(
+                                ast.Call(
+                                    func=load("__exc_clear"),
+                                    args=[],
+                                    keywords=[],
+                                    starargs=None,
+                                    kwargs=None,
+                                )
+                            )
+                        ],
+                    )],
+                )]
 
         return body
 
+    def translate_proxy(self, engine, *args):
+        """Default implementation delegates to ``translate`` method."""
+
+        return self.translate(*args)
+
     def translate(self, expression, target):
         """Return statements that assign a value to ``target``."""
 
@@ -205,7 +223,7 @@ class PathExpr(TalesExpr):
 
 
 class PythonExpr(TalesExpr):
-    """Python expression compiler.
+    r"""Python expression compiler.
 
     >>> test(PythonExpr('2 + 2'))
     4
@@ -218,7 +236,7 @@ class PythonExpr(TalesExpr):
 
     To include a pipe character, use a backslash escape sequence:
 
-    >>> test(PythonExpr('\"\|\"'))
+    >>> test(PythonExpr(r'"\|"'))
     '|'
     """
 
@@ -239,8 +257,7 @@ class PythonExpr(TalesExpr):
 
         try:
             value = self.parse(string)
-        except SyntaxError:
-            exc = sys.exc_info()[1]
+        except SyntaxError as exc:
             raise ExpressionError(exc.msg, string)
 
         # Transform attribute lookups to allow fallback to item lookup
@@ -249,7 +266,7 @@ class PythonExpr(TalesExpr):
         return [ast.Assign(targets=[target], value=value)]
 
 
-class ImportExpr(object):
+class ImportExpr:
     re_dotted = re.compile(r'^[A-Za-z.]+$')
 
     def __init__(self, expression):
@@ -262,11 +279,11 @@ class ImportExpr(object):
             RESOLVE=Symbol(resolve_dotted),
             NAME=ast.Str(s=string),
             mode="eval",
-            )
+        )
         return [ast.Assign(targets=[target], value=value)]
 
 
-class NotExpr(object):
+class NotExpr:
     """Negates the expression.
 
     >>> engine = SimpleEngine(PythonExpr)
@@ -286,7 +303,7 @@ class NotExpr(object):
         return body + template("target = not target", target=target)
 
 
-class StructureExpr(object):
+class StructureExpr:
     """Wraps the expression result as 'structure'.
 
     >>> engine = SimpleEngine(PythonExpr)
@@ -307,10 +324,10 @@ class StructureExpr(object):
             "target = wrapper(target)",
             target=target,
             wrapper=self.wrapper_class
-            )
+        )
 
 
-class IdentityExpr(object):
+class IdentityExpr:
     """Identity expression.
 
     Exists to demonstrate the interface.
@@ -327,8 +344,9 @@ class IdentityExpr(object):
         return compiler.assign_value(target)
 
 
-class StringExpr(object):
+class StringExpr:
     """Similar to the built-in ``string.Template``, but uses an
+
     expression engine to support pluggable string substitution
     expressions.
 
@@ -356,21 +374,27 @@ class StringExpr(object):
     >>> test(StringExpr('Hello $name!', True))
     'Hello $name!'
 
-    We can escape interpolation using the standard escaping
-    syntax:
+    To avoid interpolation, use two dollar symbols. Note that only a
+    single symbol will appear in the output.
 
-    >>> test(StringExpr('\\${name}'))
-    '\\\${name}'
+    >>> test(StringExpr('$${name}'))
+    '${name}'
+
+    In previous versions, it was possible to escape using a regular
+    backslash coding, but this is no longer supported.
+
+    >>> test(StringExpr(r'\\${name}'), name='Hello world!')
+    '\\\\Hello world!'
 
     Multiple interpolations in one:
 
-    >>> test(StringExpr(\"Hello ${'a'}${'b'}${'c'}!\"))
+    >>> test(StringExpr("Hello ${'a'}${'b'}${'c'}!"))
     'Hello abc!'
 
     Here's a more involved example taken from a javascript source:
 
     >>> result = test(StringExpr(\"\"\"
-    ... function(oid) {
+    ... function($$, oid) {
     ...     $('#' + oid).autocomplete({source: ${'source'}});
     ... }
     ... \"\"\"))
@@ -378,6 +402,27 @@ class StringExpr(object):
     >>> 'source: source' in result
     True
 
+    As noted previously, the double-dollar escape also affects
+    non-interpolation expressions.
+
+    >>> 'function($, oid)' in result
+    True
+
+    >>> test(StringExpr('test ${1}${2}'))
+    'test 12'
+
+    >>> test(StringExpr('test $${1}${2}'))
+    'test ${1}2'
+
+    >>> test(StringExpr('test $$'))
+    'test $'
+
+    >>> test(StringExpr('$$.ajax(...)'))
+    '$.ajax(...)'
+
+    >>> test(StringExpr('test $$ ${1}'))
+    'test $ 1'
+
     In the above examples, the expression is evaluated using the
     dummy engine which just returns the input as a string.
 
@@ -387,7 +432,7 @@ class StringExpr(object):
 
     >>> class engine:
     ...     @staticmethod
-    ...     def parse(expression):
+    ...     def parse(expression, char_escape=None):
     ...         class compiler:
     ...             @staticmethod
     ...             def assign_text(target):
@@ -422,13 +467,18 @@ class StringExpr(object):
         return self.translator(name, engine)
 
 
-class ProxyExpr(StringExpr):
-    def __init__(self, name, *args):
-        super(ProxyExpr, self).__init__(*args)
+class ProxyExpr(TalesExpr):
+    braces_required = False
+
+    def __init__(self, name, expression, ignore_prefix=True):
+        super().__init__(expression)
+        self.ignore_prefix = ignore_prefix
         self.name = name
 
-    def __call__(self, target, engine):
-        assignment = super(ProxyExpr, self).__call__(target, engine)
+    def translate_proxy(self, engine, expression, target):
+        translator = Interpolator(expression, self.braces_required)
+        assignment = translator(target, engine)
+
         return assignment + [
             ast.Assign(targets=[target], value=ast.Call(
                 func=load(self.name),
@@ -436,11 +486,11 @@ class ProxyExpr(StringExpr):
                 keywords=[],
                 starargs=None,
                 kwargs=None
-                ))
-            ]
+            ))
+        ]
 
 
-class ExistsExpr(object):
+class ExistsExpr:
     """Boolean wrapper.
 
     Return 0 if the expression results in an exception, otherwise 1.
@@ -457,32 +507,32 @@ class ExistsExpr(object):
 
     """
 
-    exceptions = AttributeError, LookupError, TypeError, NameError, KeyError
+    exceptions = AttributeError, LookupError, TypeError, NameError
 
     def __init__(self, expression):
         self.expression = expression
 
     def __call__(self, target, engine):
         ignore = store("_ignore")
-        compiler = engine.parse(self.expression)
+        compiler = engine.parse(self.expression, False)
         body = compiler.assign_value(ignore)
 
         classes = map(resolve_global, self.exceptions)
 
         return [
-            ast.TryExcept(
+            Try(
                 body=body,
                 handlers=[ast.ExceptHandler(
                     type=ast.Tuple(elts=classes, ctx=ast.Load()),
                     name=None,
                     body=template("target = 0", target=target),
-                    )],
+                )],
                 orelse=template("target = 1", target=target)
-                )
-            ]
+            )
+        ]
 
 
-class ExpressionParser(object):
+class ExpressionParser:
     def __init__(self, factories, default):
         self.factories = factories
         self.default = default
@@ -497,28 +547,27 @@ class ExpressionParser(object):
 
         try:
             factory = self.factories[prefix]
-        except KeyError:
-            exc = sys.exc_info()[1]
+        except KeyError as exc:
             raise LookupError(
                 "Unknown expression type: %s." % str(exc)
-                )
+            )
 
         return factory(expression)
 
 
-class SimpleEngine(object):
+class SimpleEngine:
     expression = PythonExpr
 
     def __init__(self, expression=None):
         if expression is not None:
             self.expression = expression
 
-    def parse(self, string):
+    def parse(self, string, handle_errors=False, char_escape=None):
         compiler = self.expression(string)
         return SimpleCompiler(compiler, self)
 
 
-class SimpleCompiler(object):
+class SimpleCompiler:
     def __init__(self, compiler, engine):
         self.compiler = compiler
         self.engine = engine
@@ -538,4 +587,4 @@ class SimpleCompiler(object):
             "target = builtin(target)",
             target=target,
             builtin=builtin
-            )
+        )
