@@ -1,71 +1,82 @@
-from __future__ import with_statement
-
+import hashlib
+import inspect
+import logging
 import os
 import sys
-import copy
-import atexit
-import hashlib
-import shutil
-import logging
 import tempfile
-import inspect
 
-pkg_digest = hashlib.sha1(__name__.encode('utf-8'))
-
-try:
-    import pkg_resources
-except ImportError:
-    logging.info("Setuptools not installed. Unable to determine version.")
-else:
-    for path in sys.path:
-        for distribution in pkg_resources.find_distributions(path):
-            if distribution.has_version():
-                version = distribution.version.encode('utf-8')
-                pkg_digest.update(version)
-
-
-from .exc import TemplateError
-from .exc import ExceptionFormatter
 from .compiler import Compiler
-from .config import DEBUG_MODE
 from .config import AUTO_RELOAD
-from .config import EAGER_PARSING
 from .config import CACHE_DIRECTORY
-from .loader import ModuleLoader
+from .config import DEBUG_MODE
+from .config import EAGER_PARSING
+from .exc import ExceptionFormatter
+from .exc import RenderError
+from .exc import TemplateError
 from .loader import MemoryLoader
+from .loader import ModuleLoader
 from .nodes import Module
 from .utils import DebuggingOutputStream
 from .utils import Scope
+from .utils import create_formatted_exception
 from .utils import join
 from .utils import mangle
-from .utils import derive_formatted_exception
-from .utils import read_bytes
 from .utils import raise_with_traceback
-from .utils import byte_string
+from .utils import read_bytes
+from .utils import value_repr
+
+
+try:
+    RecursionError
+except NameError:
+    RecursionError = RuntimeError
+
+
+def get_package_versions():
+    try:
+        import pkg_resources
+    except ImportError:
+        logging.info("Setuptools not installed. Unable to determine version.")
+        return []
+
+    versions = dict()
+    for path in sys.path:
+        for distribution in pkg_resources.find_distributions(path):
+            if distribution.has_version():
+                versions.setdefault(
+                    distribution.project_name,
+                    distribution.version,
+                )
+
+    return sorted(versions.items())
+
+
+pkg_digest = hashlib.sha1(__name__.encode('utf-8'))
+for name, version in get_package_versions():
+    pkg_digest.update(name.encode('utf-8'))
+    pkg_digest.update(version.encode('utf-8'))
 
 
 log = logging.getLogger('chameleon.template')
 
 
 def _make_module_loader():
+    remove = False
     if CACHE_DIRECTORY:
         path = CACHE_DIRECTORY
     else:
         path = tempfile.mkdtemp()
+        remove = True
 
-        @atexit.register
-        def cleanup(path=path, shutil=shutil):
-            shutil.rmtree(path)
-
-    return ModuleLoader(path)
+    return ModuleLoader(path, remove)
 
 
-class BaseTemplate(object):
+class BaseTemplate:
     """Template base class.
 
     Takes a string input which must be one of the following:
 
-    - a unicode string (or string on Python 3);
+    - a string;
     - a utf-8 encoded byte string;
     - a byte string for an XML document that defines an encoding
       in the document premamble;
@@ -113,6 +124,10 @@ class BaseTemplate(object):
     # time. When not set, this is only required at evaluation time.
     strict = True
 
+    # This should return a value string representation for exception
+    # formatting.
+    value_repr = staticmethod(value_repr)
+
     def __init__(self, body=None, **config):
         self.__dict__.update(config)
 
@@ -128,7 +143,7 @@ class BaseTemplate(object):
         return self.render(**kwargs)
 
     def __repr__(self):
-        return "<%s %s>" % (self.__class__.__name__, self.filename)
+        return "<{} {}>".format(self.__class__.__name__, self.filename)
 
     @property
     def keep_body(self):
@@ -143,10 +158,10 @@ class BaseTemplate(object):
         return self.__dict__.get('keep_source', DEBUG_MODE)
 
     def cook(self, body):
-        digest = self._digest(body)
         builtins_dict = self.builtins.copy()
         builtins_dict.update(self.extra_builtins)
-        names, builtins = zip(*builtins_dict.items())
+        names, builtins = zip(*sorted(builtins_dict.items()))
+        digest = self.digest(body, names)
         program = self._cook(body, digest, names)
 
         initialize = program['initialize']
@@ -173,30 +188,42 @@ class BaseTemplate(object):
         stream = self.output_stream_factory()
         try:
             self._render(stream, econtext, rcontext)
-        except:
-            cls, exc, tb = sys.exc_info()
-            errors = rcontext.get('__error__')
-            if errors:
-                try:
-                    exc = copy.copy(exc)
-                except TypeError:
-                    name = type(exc).__name__
-                    log.warn("Unable to copy exception of type '%s'." % name)
-                else:
-                    formatter = ExceptionFormatter(errors, econtext, rcontext)
-                    exc = derive_formatted_exception(exc, cls, formatter)
-
-                raise_with_traceback(exc, tb)
-
+        except RecursionError:
             raise
+        except BaseException:
+            cls, exc, tb = sys.exc_info()
+            try:
+                errors = rcontext.get('__error__')
+                if errors:
+                    formatter = exc.__str__
+                    if isinstance(formatter, ExceptionFormatter):
+                        if errors is not formatter._errors:
+                            formatter._errors.extend(errors)
+                        raise
+
+                    formatter = ExceptionFormatter(
+                        errors, econtext, rcontext, self.value_repr)
+
+                    try:
+                        exc = create_formatted_exception(
+                            exc, cls, formatter, RenderError
+                        )
+                    except TypeError:
+                        pass
+
+                    raise_with_traceback(exc, tb)
+
+                raise
+            finally:
+                del exc, tb
 
         return join(stream)
 
     def write(self, body):
-        if isinstance(body, byte_string):
+        if isinstance(body, bytes):
             body, encoding, content_type = read_bytes(
                 body, self.default_encoding
-                )
+            )
         else:
             content_type = body.startswith('<?xml')
             encoding = None
@@ -206,23 +233,23 @@ class BaseTemplate(object):
 
         self.cook(body)
 
-    def _get_module_name(self, digest):
-        return "%s.py" % digest
+    def _get_module_name(self, name):
+        return "%s.py" % name
 
-    def _cook(self, body, digest, builtins):
-        name = self._get_module_name(digest)
-        cooked = self.loader.get(name)
+    def _cook(self, body, name, builtins):
+        filename = self._get_module_name(name)
+        cooked = self.loader.get(filename)
         if cooked is None:
             try:
-                source = self._make(body, builtins)
+                source = self._compile(body, builtins)
                 if self.debug:
-                    source = "# template: %s\n#\n%s" % (self.filename, source)
+                    source = "# template: {}\n#\n{}".format(
+                        self.filename, source)
                 if self.keep_source:
                     self.source = source
-                cooked = self.loader.build(source, name)
-            except TemplateError:
-                exc = sys.exc_info()[1]
-                exc.filename = self.filename
+                cooked = self.loader.build(source, filename)
+            except TemplateError as exc:
+                exc.token.filename = self.filename
                 raise
         elif self.keep_source:
             module = sys.modules.get(cooked.get('__name__'))
@@ -230,24 +257,28 @@ class BaseTemplate(object):
                 self.source = inspect.getsource(module)
             else:
                 self.source = None
-
         return cooked
 
-    def _digest(self, body):
+    def digest(self, body, names):
         class_name = type(self).__name__.encode('utf-8')
         sha = pkg_digest.copy()
         sha.update(body.encode('utf-8', 'ignore'))
         sha.update(class_name)
-        return sha.hexdigest()
+        digest = sha.hexdigest()
 
-    def _compile(self, program, builtins):
-        compiler = Compiler(self.engine, program, builtins, strict=self.strict)
-        return compiler.code
+        if self.filename and self.filename is not BaseTemplate.filename:
+            digest = os.path.splitext(self.filename)[0] + '-' + digest
 
-    def _make(self, body, builtins):
+        return digest
+
+    def _compile(self, body, builtins):
         program = self.parse(body)
         module = Module("initialize", program)
-        return self._compile(module, builtins)
+        compiler = Compiler(
+            self.engine, module, self.filename, body,
+            builtins, strict=self.strict
+        )
+        return compiler.code
 
 
 class BaseTemplateFile(BaseTemplate):
@@ -261,11 +292,16 @@ class BaseTemplateFile(BaseTemplate):
     # performance hit
     auto_reload = AUTO_RELOAD
 
-    def __init__(self, filename, auto_reload=None, **config):
+    def __init__(
+            self,
+            filename,
+            auto_reload=None,
+            post_init_hook=None,
+            **config):
         # Normalize filename
         filename = os.path.abspath(
             os.path.normpath(os.path.expanduser(filename))
-            )
+        )
 
         self.filename = filename
 
@@ -273,7 +309,10 @@ class BaseTemplateFile(BaseTemplate):
         if auto_reload is not None:
             self.auto_reload = auto_reload
 
-        super(BaseTemplateFile, self).__init__(**config)
+        super().__init__(**config)
+
+        if post_init_hook is not None:
+            post_init_hook()
 
         if EAGER_PARSING:
             self.cook_check()
@@ -288,12 +327,13 @@ class BaseTemplateFile(BaseTemplate):
 
         if self._cooked is False:
             body = self.read()
+            log.debug("cooking %r (%d bytes)..." % (self.filename, len(body)))
             self.cook(body)
 
     def mtime(self):
         try:
             return os.path.getmtime(self.filename)
-        except (IOError, OSError):
+        except OSError:
             return 0
 
     def read(self):
@@ -302,7 +342,7 @@ class BaseTemplateFile(BaseTemplate):
 
         body, encoding, content_type = read_bytes(
             data, self.default_encoding
-            )
+        )
 
         # In non-XML mode, we support various platform-specific line
         # endings and convert them to the UNIX newline character
@@ -314,10 +354,10 @@ class BaseTemplateFile(BaseTemplate):
 
         return body
 
-    def _get_module_name(self, digest):
+    def _get_module_name(self, name):
         filename = os.path.basename(self.filename)
         mangled = mangle(filename)
-        return "%s_%s.py" % (mangled, digest)
+        return "{}_{}.py".format(mangled, name)
 
     def _get_filename(self):
         return self.__dict__.get('filename')
