@@ -1,72 +1,35 @@
+import codecs
+import logging
 import os
 import re
-import sys
-import codecs
-
-version = sys.version_info[:3]
-
-try:
-    import ast as _ast
-except ImportError:
-    from chameleon import ast24 as _ast
+from html import entities as htmlentitydefs
 
 
-class ASTProxy(object):
-    aliases = {
-        # Python 3.3
-        'TryExcept': 'Try',
-        'TryFinally': 'Try',
-        }
-
-    def __getattr__(self, name):
-        return _ast.__dict__.get(name) or getattr(_ast, self.aliases[name])
+log = logging.getLogger('chameleon.utils')
+marker = object()
 
 
-ast = ASTProxy()
+def safe_native(s, encoding='utf-8'):
+    if not isinstance(s, str):
+        s = bytes.decode(s, encoding, 'replace')
+    return s
 
 
-# Python 2
-if version < (3, 0, 0):
-    import htmlentitydefs
-    import __builtin__ as builtins
+def raise_with_traceback(exc, tb):
+    exc.__traceback__ = tb
+    raise exc
 
-    from .py25 import raise_with_traceback
 
-    chr = unichr
-    native_string = str
-    decode_string = unicode
-    encode_string = str
-    unicode_string = unicode
-    string_type = basestring
-    byte_string = str
+def encode_string(s):
+    return bytes(s, 'utf-8')
 
-    def safe_native(s, encoding='utf-8'):
-        if not isinstance(s, unicode):
-            s = decode_string(s, encoding, 'replace')
 
-        return s.encode(encoding)
-
-# Python 3
-else:
-    from html import entities as htmlentitydefs
-    import builtins
-
-    byte_string = bytes
-    string_type = str
-    native_string = str
-    decode_string = bytes.decode
-    encode_string = lambda s: bytes(s, 'utf-8')
-    unicode_string = str
-
-    def safe_native(s, encoding='utf-8'):
-        if not isinstance(s, str):
-            s = decode_string(s, encoding, 'replace')
-
-        return s
-
-    def raise_with_traceback(exc, tb):
-        exc.__traceback__ = tb
-        raise exc
+def text_(s, encoding='latin-1', errors='strict'):
+    """ If ``s`` is an instance of ``bytes``, return
+    ``s.decode(encoding, errors)``, otherwise return ``s``"""
+    if isinstance(s, bytes):
+        return s.decode(encoding, errors)
+    return s
 
 
 entity_re = re.compile(r'&(#?)(x?)(\d{1,5}|\w{1,8});')
@@ -81,7 +44,7 @@ xml_prefixes = (
     (codecs.BOM_UTF32_BE, 'utf-32-be'),
     (codecs.BOM_UTF32_LE, 'utf-32-le'),
     (codecs.BOM_UTF32, 'utf-32'),
-    )
+)
 
 
 def _has_encoding(encoding):
@@ -95,10 +58,10 @@ def _has_encoding(encoding):
 
 # Precomputed prefix table
 _xml_prefixes = tuple(
-    (bom, str('<?xml').encode(encoding), encoding)
+    (bom, '<?xml'.encode(encoding), encoding)
     for bom, encoding in reversed(xml_prefixes)
     if _has_encoding(encoding)
-    )
+)
 
 _xml_decl = encode_string("<?xml")
 
@@ -106,12 +69,12 @@ RE_META = re.compile(
     r'\s*<meta\s+http-equiv=["\']?Content-Type["\']?'
     r'\s+content=["\']?([^;]+);\s*charset=([^"\']+)["\']?\s*/?\s*>\s*',
     re.IGNORECASE
-    )
+)
 
 RE_ENCODING = re.compile(
-    r'encoding\s*=\s*(?:"|\')(?P<encoding>[\w\-]+)(?:"|\')'.encode('ascii'),
+    br'encoding\s*=\s*(?:"|\')(?P<encoding>[\w\-]+)(?:"|\')',
     re.IGNORECASE
-    )
+)
 
 
 def read_encoded(data):
@@ -123,9 +86,9 @@ def read_bytes(body, default_encoding):
         if body.startswith(bom):
             document = body.decode(encoding)
             return document, encoding, \
-                   "text/xml" if document.startswith("<?xml") else None
+                "text/xml" if document.startswith("<?xml") else None
 
-        if prefix != '<?xml' and body.startswith(prefix):
+        if prefix != encode_string('<?xml') and body.startswith(prefix):
             return body.decode(encoding), encoding, "text/xml"
 
     if body.startswith(_xml_decl):
@@ -150,7 +113,7 @@ def detect_encoding(body, default_encoding):
 
 
 def read_xml_encoding(body):
-    if body.startswith('<?xml'.encode('ascii')):
+    if body.startswith(b'<?xml'):
         match = RE_ENCODING.search(body)
         if match is not None:
             return match.group('encoding').decode('ascii')
@@ -197,18 +160,31 @@ def substitute_entity(match, n2cp=htmlentitydefs.name2codepoint):
             return match.group()
 
 
-def derive_formatted_exception(exc, cls, formatter):
+def create_formatted_exception(exc, cls, formatter, base=Exception):
     try:
-        new = type(cls.__name__, (cls, Exception), {
-            '__str__': formatter,
+        try:
+            new = type(cls.__name__, (cls, base), {
+                '__str__': formatter,
+                '_original__str__': exc.__str__,
+                '__new__': BaseException.__new__,
+                '__module__': cls.__module__,
             })
-        exc.__class__ = new
-    except TypeError:
-        d = exc.__dict__
-        exc = cls.__new__(new)
-        exc.__dict__ = d
+        except TypeError:
+            new = cls
 
-    return exc
+        try:
+            inst = BaseException.__new__(new)
+        except TypeError:
+            inst = cls.__new__(new)
+
+        BaseException.__init__(inst, *exc.args)
+        inst.__dict__ = exc.__dict__
+
+        return inst
+    except ValueError:
+        name = type(exc).__name__
+        log.warn("Unable to copy exception of type '%s'." % name)
+        raise TypeError(exc)
 
 
 def unescape(string):
@@ -225,23 +201,23 @@ _concat = "".join
 def join(stream):
     """Concatenate stream.
 
-    >>> join(('Hello', ' ', 'world'))
-    'Hello world'
+    >>> print(join(('Hello', ' ', 'world')))
+    Hello world
 
     >>> join(('Hello', 0))
     Traceback (most recent call last):
      ...
-    TypeError: ... int ...
+    TypeError: ... expected ...
 
     """
 
     try:
         return _concat(stream)
-    except:
+    except BaseException:
         # Loop through stream and coerce each element into unicode;
         # this should raise an exception
         for element in stream:
-            unicode_string(element)
+            str(element)
 
         # In case it didn't, re-raise the original exception
         raise
@@ -249,7 +225,7 @@ def join(stream):
 
 def decode_htmlentities(string):
     """
-    >>> native_string(decode_htmlentities('&amp;amp;'))
+    >>> str(decode_htmlentities('&amp;amp;'))
     '&amp;'
 
     """
@@ -287,7 +263,7 @@ def _resolve_dotted(name, module=None):
 
 
 def resolve_dotted(dotted):
-    if not dotted in module_cache:
+    if dotted not in module_cache:
         resolved = _resolve_dotted(dotted)
         module_cache[dotted] = resolved
     return module_cache[dotted]
@@ -300,24 +276,22 @@ def limit_string(s, max_length=53):
     return s
 
 
-def format_kwargs(kwargs):
-    items = []
-    for name, value in kwargs.items():
-        if isinstance(value, string_type):
-            short = limit_string(value)
-            items.append((name, short.replace('\n', '\\n')))
-        elif isinstance(value, (int, float)):
-            items.append((name, value))
-        elif isinstance(value, dict):
-            items.append((name, '{...} (%d)' % len(value)))
-        else:
-            items.append((name,
-                "<%s %s at %s>" % (
-                    type(value).__name__,
-                    getattr(value, '__name__', "-"),
-                    hex(abs(id(value))))))
+def value_repr(value):
+    if isinstance(value, str):
+        short = limit_string(value)
+        return short.replace('\n', '\\n')
+    if isinstance(value, (int, float)):
+        return value
+    if isinstance(value, dict):
+        return '{...} (%d)' % len(value)
 
-    return ["%s: %s" % item for item in items]
+    try:
+        name = str(getattr(value, '__name__', None)),
+    except:  # noqa: E722 do not use bare 'except'
+        name = '-'
+
+    return '<{} {} at {}>'.format(
+        type(value).__name__, name, hex(abs(id(value))))
 
 
 class callablestr(str):
@@ -334,7 +308,7 @@ class callableint(int):
         return self
 
 
-class descriptorstr(object):
+class descriptorstr:
     __slots__ = "function", "__name__"
 
     def __init__(self, function):
@@ -345,7 +319,7 @@ class descriptorstr(object):
         return callablestr(self.function(context))
 
 
-class descriptorint(object):
+class descriptorint:
     __slots__ = "function", "__name__"
 
     def __init__(self, function):
@@ -358,28 +332,79 @@ class descriptorint(object):
 
 class DebuggingOutputStream(list):
     def append(self, value):
-        if not isinstance(value, string_type):
+        if not isinstance(value, str):
             raise TypeError(value)
 
-        unicode_string(value)
+        str(value)
         list.append(self, value)
 
 
 class Scope(dict):
-    set_local = setLocal = dict.__setitem__
+    """
+    >>> scope = Scope()
+    >>> scope['a'] = 1
+    >>> copy = scope.copy()
 
-    __slots__ = "set_global",
+    Setting a local value and then a global value, we expect the local value
+    to take precedence.
 
-    def __new__(cls, *args):
-        inst = dict.__new__(cls, *args)
-        inst.set_global = inst.__setitem__
-        return inst
+    >>> copy['a'] = 2
+    >>> copy.set_global('a', 3)
+    >>> assert copy['a'] == 2
+
+    However, setting a new global value should be immediately visible.
+
+    >>> copy.set_global('b', 1)
+    >>> assert copy['b'] == 1
+
+    Make sure the objects are reference-counted, not requiring a full
+    collection to be disposed of.
+
+    >>> import gc
+    >>> _ = gc.collect()
+    >>> del copy
+    >>> del scope
+    >>> import platform
+    >>> assert gc.collect() == (
+    ...     0 if platform.python_implementation() == 'CPython' else None
+    ... )
+    """
+
+    __slots__ = "_root",
+
+    set_local = dict.__setitem__
 
     def __getitem__(self, key):
-        try:
-            return dict.__getitem__(self, key)
-        except KeyError:
-            raise NameError(key)
+        value = self.get(key, marker)
+        if value is marker:
+            raise KeyError(key)
+        return value
+
+    def __contains__(self, key):
+        return self.get(key, marker) is not marker
+
+    def get(self, key, default=None):
+        value = dict.get(self, key, marker)
+        if value is not marker:
+            return value
+
+        root = getattr(self, "_root", marker)
+        if root is not marker:
+            value = dict.get(root, key, marker)
+
+            if value is not marker:
+                return value
+
+        return default
+
+    def items(self):
+        raise NotImplementedError()
+
+    def keys(self):
+        raise NotImplementedError()
+
+    def values(self):
+        raise NotImplementedError()
 
     @property
     def vars(self):
@@ -387,21 +412,68 @@ class Scope(dict):
 
     def copy(self):
         inst = Scope(self)
-        inst.set_global = self.set_global
+        root = getattr(self, "_root", self)
+        inst._root = root
         return inst
 
+    def set_global(self, name, value):
+        root = getattr(self, "_root", self)
+        root[name] = value
 
-class ListDictProxy(object):
-    def __init__(self, l):
-        self._l = l
+    def get_name(self, key):
+        value = self.get(key, marker)
+        if value is marker:
+            raise NameError(key)
+        return value
+
+    setLocal = set_local
+    setGlobal = set_global
+
+
+class ListDictProxy:
+    def __init__(self, _l):
+        self._l = _l
 
     def get(self, key):
         return self._l[-1].get(key)
 
 
-class Markup(unicode_string):
+class Markup(str):
+    """Wraps a string to always render as structure.
+
+    >>> Markup('<br />')
+    s'<br />'
+    """
+
     def __html__(self):
-        return unicode_string(self)
+        return str(self)
 
     def __repr__(self):
-        return "s'%s'" % self.s
+        return "s'%s'" % self
+
+
+class ImportableMarker:
+    def __init__(self, module, name):
+        self.__module__ = module
+        self.name = name
+
+    @property
+    def __name__(self):
+        return "%s_MARKER" % self.name
+
+    def __repr__(self):
+        return '<%s>' % self.name
+
+
+def lookup_attr(obj, key):
+    try:
+        return getattr(obj, key)
+    except AttributeError as exc:
+        try:
+            get = obj.__getitem__
+        except AttributeError:
+            raise exc
+        try:
+            return get(key)
+        except KeyError:
+            raise exc
