@@ -1,41 +1,36 @@
 import re
 
-try:
-    import ast
-except ImportError:
-    from chameleon import ast24 as ast
 
 try:
     str = unicode
 except NameError:
     long = int
 
+import ast
+from copy import copy
 from functools import partial
 
-from ..program import ElementProgram
-
-from ..namespaces import XML_NS
-from ..namespaces import XMLNS_NS
-from ..namespaces import I18N_NS as I18N
-from ..namespaces import TAL_NS as TAL
-from ..namespaces import METAL_NS as METAL
-from ..namespaces import META_NS as META
-
-from ..astutil import Static
-from ..astutil import parse
-from ..astutil import marker
-
-from .. import tal
-from .. import metal
 from .. import i18n
+from .. import metal
 from .. import nodes
-
+from .. import tal
+from ..astutil import Static
+from ..astutil import Symbol
+from ..astutil import parse
+from ..exc import CompilationError
 from ..exc import LanguageError
 from ..exc import ParseError
-from ..exc import CompilationError
-
+from ..namespaces import I18N_NS as I18N
+from ..namespaces import META_NS as META
+from ..namespaces import METAL_NS as METAL
+from ..namespaces import TAL_NS as TAL
+from ..namespaces import XML_NS
+from ..namespaces import XMLNS_NS
+from ..program import ElementProgram
+from ..utils import ImportableMarker
 from ..utils import decode_htmlentities
 
+
 try:
     str = unicode
 except NameError:
@@ -44,6 +39,11 @@ except NameError:
 
 missing = object()
 
+re_trim = re.compile(r'($\s+|\s+^)', re.MULTILINE)
+
+EMPTY_DICT = Static(ast.Dict(keys=[], values=[]))
+CANCEL_MARKER = ImportableMarker(__name__, "CANCEL")
+
 
 def skip(node):
     return node
@@ -60,7 +60,21 @@ def validate_attributes(attributes, namespace, whitelist):
         if ns == namespace and name not in whitelist:
             raise CompilationError(
                 "Bad attribute for namespace '%s'" % ns, name
-                )
+            )
+
+
+def convert_data_attributes(ns_attrs, attrs, namespaces):
+    d = 0
+    for i, attr in list(enumerate(attrs)):
+        name = attr['name']
+        if name.startswith('data-'):
+            name = name[5:]
+            if '-' not in name:
+                continue
+            prefix, name = name.split('-', 1)
+            ns_attrs[namespaces[prefix], name] = attr['value']
+            attrs.pop(i - d)
+            d += 1
 
 
 class MacroProgram(ElementProgram):
@@ -73,7 +87,7 @@ class MacroProgram(ElementProgram):
         'metal': METAL,
         'i18n': I18N,
         'meta': META,
-        }
+    }
 
     DROP_NS = TAL, METAL, I18N, META
 
@@ -83,6 +97,7 @@ class MacroProgram(ElementProgram):
     _interpolation_enabled = True
     _whitespace = "\n"
     _last = ""
+    _cancel_marker = Symbol(CANCEL_MARKER)
 
     # Macro name (always trivial for a macro program)
     name = None
@@ -113,6 +128,17 @@ class MacroProgram(ElementProgram):
     # If set, additional attribute whitespace will be stripped.
     trim_attribute_space = False
 
+    # If set, data attributes can be used instead of namespace
+    # attributes, e.g. "data-tal-content" instead of "tal:content".
+    enable_data_attributes = False
+
+    # If set, XML namespaces are restricted to the list of those
+    # defined and used by the page template language.
+    restricted_namespace = True
+
+    # If set, expression interpolation is enabled in comments.
+    enable_comment_interpolation = True
+
     def __init__(self, *args, **kwargs):
         # Internal array for switch statements
         self._switches = []
@@ -135,9 +161,12 @@ class MacroProgram(ElementProgram):
             'implicit_i18n_translate',
             'implicit_i18n_attributes',
             'trim_attribute_space',
-            )
+            'enable_data_attributes',
+            'enable_comment_interpolation',
+            'restricted_namespace',
+        )
 
-        super(MacroProgram, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     @property
     def macros(self):
@@ -147,16 +176,21 @@ class MacroProgram(ElementProgram):
         return tuple(
             nodes.Macro(name, [nodes.Context(node)])
             for name, node in macros
-            )
+        )
 
     def visit_default(self, node):
         return nodes.Text(node)
 
     def visit_element(self, start, end, children):
         ns = start['ns_attrs']
+        attrs = start['attrs']
+
+        if self.enable_data_attributes:
+            attrs = list(attrs)
+            convert_data_attributes(ns, attrs, start['ns_map'])
 
         for (prefix, attr), encoded in tuple(ns.items()):
-            if prefix == TAL:
+            if prefix == TAL or prefix == METAL:
                 ns[prefix, attr] = decode_htmlentities(encoded)
 
         # Validate namespace attributes
@@ -180,7 +214,8 @@ class MacroProgram(ElementProgram):
         except KeyError:
             switch = None
         else:
-            switch = nodes.Value(clause)
+            value = nodes.Value(clause)
+            switch = value
 
         self._switches.append(switch)
 
@@ -190,17 +225,31 @@ class MacroProgram(ElementProgram):
         use_macro = ns.get((METAL, 'use-macro'))
         extend_macro = ns.get((METAL, 'extend-macro'))
         if use_macro or extend_macro:
+            omit = True
             slots = []
             self._use_macro.append(slots)
 
             if use_macro:
                 inner = nodes.UseExternalMacro(
                     nodes.Value(use_macro), slots, False
-                    )
+                )
+                macro_name = use_macro
             else:
                 inner = nodes.UseExternalMacro(
                     nodes.Value(extend_macro), slots, True
-                    )
+                )
+                macro_name = extend_macro
+
+            # While the macro executes, it should have access to the name it
+            # was called with as 'macroname'. Splitting on / mirrors zope.tal
+            # and is a concession to the path expression syntax.
+            macro_name = macro_name.rsplit('/', 1)[-1]
+            inner = nodes.Define(
+                [nodes.Assignment(
+                    ["macroname"], Static(ast.Str(macro_name)), True)],
+                inner,
+            )
+            STATIC_ATTRIBUTES = None
         # -or- include tag
         else:
             content = nodes.Sequence(body)
@@ -212,12 +261,14 @@ class MacroProgram(ElementProgram):
                 pass
             else:
                 key, value = tal.parse_substitution(clause)
-                xlate = True if ns.get((I18N, 'translate')) == '' else False
-                content = self._make_content_node(value, content, key, xlate)
+                translate = ns.get((I18N, 'translate')) == ''
+                content = self._make_content_node(
+                    value, content, key, translate,
+                )
 
                 if end is None:
                     # Make sure start-tag has opening suffix.
-                    start['suffix']  = ">"
+                    start['suffix'] = ">"
 
                     # Explicitly set end-tag.
                     end = {
@@ -225,7 +276,7 @@ class MacroProgram(ElementProgram):
                         'name': start['name'],
                         'space': '',
                         'suffix': '>'
-                        }
+                    }
 
             # i18n:translate
             try:
@@ -242,7 +293,7 @@ class MacroProgram(ElementProgram):
             try:
                 clause = ns[TAL, 'attributes']
             except KeyError:
-                TAL_ATTRIBUTES = {}
+                TAL_ATTRIBUTES = []
             else:
                 TAL_ATTRIBUTES = tal.parse_attributes(clause)
 
@@ -256,15 +307,15 @@ class MacroProgram(ElementProgram):
 
             # Prepare attributes from TAL language
             prepared = tal.prepare_attributes(
-                start['attrs'], TAL_ATTRIBUTES,
+                attrs, TAL_ATTRIBUTES,
                 I18N_ATTRIBUTES, ns, self.DROP_NS
-                )
+            )
 
             # Create attribute nodes
             STATIC_ATTRIBUTES = self._create_static_attributes(prepared)
             ATTRIBUTES = self._create_attributes_nodes(
-                prepared, I18N_ATTRIBUTES
-                )
+                prepared, I18N_ATTRIBUTES, STATIC_ATTRIBUTES,
+            )
 
             # Start- and end nodes
             start_tag = nodes.Start(
@@ -272,14 +323,14 @@ class MacroProgram(ElementProgram):
                 self._maybe_trim(start['prefix']),
                 self._maybe_trim(start['suffix']),
                 ATTRIBUTES
-                )
+            )
 
             end_tag = nodes.End(
                 end['name'],
                 end['space'],
                 self._maybe_trim(end['prefix']),
                 self._maybe_trim(end['suffix']),
-                ) if end is not None else None
+            ) if end is not None else None
 
             # tal:omit-tag
             try:
@@ -308,13 +359,7 @@ class MacroProgram(ElementProgram):
                     start_tag,
                     end_tag,
                     content,
-                    )
-
-                # Assign static attributes dictionary to "attrs" value
-                inner = nodes.Define(
-                    [nodes.Alias(["attrs"], STATIC_ATTRIBUTES)],
-                    inner,
-                    )
+                )
 
                 if omit is not False:
                     inner = nodes.Cache([omit], inner)
@@ -326,8 +371,10 @@ class MacroProgram(ElementProgram):
                 pass
             else:
                 key, value = tal.parse_substitution(clause)
-                xlate = True if ns.get((I18N, 'translate')) == '' else False
-                inner = self._make_content_node(value, inner, key, xlate)
+                translate = ns.get((I18N, 'translate')) == ''
+                inner = self._make_content_node(
+                    value, inner, key, translate
+                )
 
         # metal:define-slot
         try:
@@ -341,18 +388,39 @@ class MacroProgram(ElementProgram):
         try:
             clause = ns[TAL, 'define']
         except KeyError:
-            DEFINE = skip
+            defines = []
         else:
             defines = tal.parse_defines(clause)
+
             if defines is None:
                 raise ParseError("Invalid define syntax.", clause)
 
-            DEFINE = partial(
-                nodes.Define,
-                [nodes.Assignment(
-                    names, nodes.Value(expr), context == "local")
-                 for (context, names, expr) in defines],
-                )
+        # i18n:target
+        try:
+            target_language = ns[I18N, 'target']
+        except KeyError:
+            pass
+        else:
+            # The name "default" is an alias for the target language
+            # variable. We simply replace it.
+            target_language = target_language.replace(
+                'default', 'target_language'
+            )
+
+            defines.append(
+                ('local', ("target_language", ), target_language)
+            )
+
+        assignments = [
+            nodes.Assignment(
+                names, nodes.Value(expr), context == "local")
+            for (context, names, expr) in defines
+        ]
+
+        # Assign static attributes dictionary to "attrs" value
+        assignments.insert(
+            0, nodes.Alias(["attrs"], STATIC_ATTRIBUTES or EMPTY_DICT))
+        DEFINE = partial(nodes.Define, assignments)
 
         # tal:case
         try:
@@ -367,16 +435,24 @@ class MacroProgram(ElementProgram):
             else:
                 raise LanguageError(
                     "Must define switch on a parent element.", clause
-                    )
-
-            CASE = lambda node: nodes.Define(
-                [nodes.Assignment(["default"], switch, True)],
-                nodes.Condition(
-                    nodes.Equality(switch, value),
-                    node,
-                    )
                 )
 
+            def CASE(node):
+                return nodes.Define(
+                    [nodes.Alias(["default"], self.default_marker)],
+                    nodes.Condition(
+                        nodes.And([
+                            nodes.BinOp(
+                                switch, nodes.IsNot, self._cancel_marker),
+                            nodes.Or([
+                                nodes.BinOp(value, nodes.Equals, switch),
+                                nodes.BinOp(
+                                    value, nodes.Equals, self.default_marker)
+                            ])
+                        ]),
+                        nodes.Cancel([switch], node, self._cancel_marker),
+                    ))
+
         # tal:repeat
         try:
             clause = ns[TAL, 'repeat']
@@ -389,13 +465,18 @@ class MacroProgram(ElementProgram):
 
             expression = nodes.Value(expr)
 
+            if start['namespace'] == TAL:
+                self._last = None
+                self._whitespace = whitespace.lstrip('\n')
+                whitespace = ""
+
             REPEAT = partial(
                 nodes.Repeat,
                 names,
                 expression,
                 context == "local",
                 whitespace
-                )
+            )
 
         # tal:condition
         try:
@@ -420,13 +501,24 @@ class MacroProgram(ElementProgram):
         else:
             DOMAIN = partial(nodes.Domain, clause)
 
+        # i18n:context
+        try:
+            clause = ns[I18N, 'context']
+        except KeyError:
+            CONTEXT = skip
+        else:
+            CONTEXT = partial(nodes.TxContext, clause)
+
         # i18n:name
         try:
             clause = ns[I18N, 'name']
         except KeyError:
             NAME = skip
         else:
-            NAME = partial(nodes.Name, clause)
+            if not clause.strip():
+                NAME = skip
+            else:
+                NAME = partial(nodes.Name, clause)
 
         # The "slot" node next is the first node level that can serve
         # as a macro slot
@@ -439,7 +531,8 @@ class MacroProgram(ElementProgram):
             REPEAT,
             SWITCH,
             DOMAIN,
-            )
+            CONTEXT,
+        )
 
         # metal:fill-slot
         try:
@@ -447,6 +540,12 @@ class MacroProgram(ElementProgram):
         except KeyError:
             pass
         else:
+            if not clause.strip():
+                raise LanguageError(
+                    "Must provide a non-trivial string for metal:fill-slot.",
+                    clause
+                )
+
             index = -(1 + int(bool(use_macro or extend_macro)))
 
             try:
@@ -455,7 +554,7 @@ class MacroProgram(ElementProgram):
                 raise LanguageError(
                     "Cannot use metal:fill-slot without metal:use-macro.",
                     clause
-                    )
+                )
 
             slots = self._use_macro[index]
             slots.append(nodes.FillSlot(clause, slot))
@@ -466,13 +565,20 @@ class MacroProgram(ElementProgram):
         except KeyError:
             pass
         else:
+            if ns.get((METAL, 'fill-slot')) is not None:
+                raise LanguageError(
+                    "Can't have 'fill-slot' and 'define-macro' "
+                    "on the same element.",
+                    clause
+                )
+
             self._macros[clause] = slot
             slot = nodes.UseInternalMacro(clause)
 
         slot = wrap(
             slot,
             NAME
-            )
+        )
 
         # tal:on-error
         try:
@@ -481,15 +587,34 @@ class MacroProgram(ElementProgram):
             ON_ERROR = skip
         else:
             key, value = tal.parse_substitution(clause)
-            translate = True if ns.get((I18N, 'translate')) == '' else False
-
-            fallback = self._make_content_node(value, None, key, translate)
+            translate = ns.get((I18N, 'translate')) == ''
+            fallback = self._make_content_node(
+                value, None, key, translate,
+            )
 
             if omit is False and start['namespace'] not in self.DROP_NS:
+                start_tag = copy(start_tag)
+
+                start_tag.attributes = nodes.Sequence(
+                    start_tag.attributes.extract(
+                        lambda attribute:
+                        isinstance(attribute, nodes.Attribute) and
+                        isinstance(attribute.expression, ast.Str)
+                    )
+                )
+
+                if end_tag is None:
+                    # Make sure start-tag has opening suffix. We don't
+                    # allow self-closing element here.
+                    start_tag.suffix = ">"
+
+                    # Explicitly set end-tag.
+                    end_tag = nodes.End(start_tag.name, '', '</', '>',)
+
                 fallback = nodes.Element(
-                start_tag,
-                end_tag,
-                fallback,
+                    start_tag,
+                    end_tag,
+                    fallback,
                 )
 
             ON_ERROR = partial(nodes.OnError, fallback, 'error')
@@ -519,13 +644,13 @@ class MacroProgram(ElementProgram):
         return wrap(
             slot,
             ON_ERROR
-            )
+        )
 
     def visit_start_tag(self, start):
         return self.visit_element(start, None, [])
 
     def visit_cdata(self, node):
-        if not self._interpolation[-1] or not '${' in node:
+        if not self._interpolation[-1] or '${' not in node:
             return nodes.Text(node)
 
         expr = nodes.Substitution(node, ())
@@ -535,10 +660,13 @@ class MacroProgram(ElementProgram):
         if node.startswith('<!--!'):
             return
 
+        if not self.enable_comment_interpolation:
+            return nodes.Text(node)
+
         if node.startswith('<!--?'):
             return nodes.Text('<!--' + node.lstrip('<!-?'))
 
-        if not self._interpolation[-1] or not '${' in node:
+        if not self._interpolation[-1] or '${' not in node:
             return nodes.Text(node)
 
         char_escape = ('&', '<', '>') if self.escape else ()
@@ -550,6 +678,13 @@ class MacroProgram(ElementProgram):
              nodes.Text(node[-3:])
              ])
 
+    def visit_processing_instruction(self, node):
+        if node['name'] != 'python':
+            text = '<?' + node['name'] + node['text'] + '?>'
+            return self.visit_text(text)
+
+        return nodes.CodeBlock(node['text'])
+
     def visit_text(self, node):
         self._last = node
 
@@ -560,38 +695,28 @@ class MacroProgram(ElementProgram):
             expression = nodes.Substitution(node, char_escape)
             return nodes.Interpolation(expression, True, translation)
 
+        node = node.replace('$$', '$')
+
         if not translation:
             return nodes.Text(node)
 
-        seq = []
-        while node:
-            m = re.search('\s+', node)
-            if m is not None:
-                s = m.start()
-                if s:
-                    t = node[:s]
-                    seq.append(nodes.Translate(t, nodes.Text(t)))
-
-                e = m.end()
-                whitespace = nodes.Text(node[s:e])
-                seq.append(whitespace)
-
-                if e < len(node):
-                    node = node[e:]
-                    continue
-
-            else:
-                seq.append(nodes.Translate(node, nodes.Text(node)))
-
-            break
-
-        return nodes.Sequence(seq)
+        match = re.search(r'(\s*)(.*\S)(\s*)', node, flags=re.DOTALL)
+        if match is not None:
+            prefix, text, suffix = match.groups()
+            normalized = re.sub(r'\s+', ' ', text)
+            return nodes.Sequence([
+                nodes.Text(prefix),
+                nodes.Translate(normalized, nodes.Text(normalized), None),
+                nodes.Text(suffix),
+            ])
+        else:
+            return nodes.Text(node)
 
     def _pop_defaults(self, kwargs, *attributes):
         for attribute in attributes:
-            default = getattr(self, attribute)
-            value = kwargs.pop(attribute, default)
-            setattr(self, attribute, value)
+            value = kwargs.pop(attribute, None)
+            if value is not None:
+                setattr(self, attribute, value)
 
     def _check_attributes(self, namespace, ns):
         if namespace in self.DROP_NS and ns.get((TAL, 'attributes')):
@@ -599,7 +724,7 @@ class MacroProgram(ElementProgram):
                 "Dynamic attributes not allowed on elements of "
                 "the namespace: %s." % namespace,
                 ns[TAL, 'attributes'],
-                )
+            )
 
         script = ns.get((TAL, 'script'))
         if script is not None:
@@ -611,13 +736,13 @@ class MacroProgram(ElementProgram):
             raise LanguageError(
                 "You cannot use tal:content and tal:replace at the same time.",
                 tal_content
-                )
+            )
 
         if tal_content and ns.get((I18N, 'translate')):
             raise LanguageError(
                 "You cannot use tal:content with non-trivial i18n:translate.",
                 tal_content
-                )
+            )
 
     def _make_content_node(self, expression, default, key, translate):
         value = nodes.Value(expression)
@@ -626,96 +751,141 @@ class MacroProgram(ElementProgram):
 
         if default is not None:
             content = nodes.Condition(
-                nodes.Identity(value, marker("default")),
+                nodes.BinOp(value, nodes.Is, self.default_marker),
                 default,
                 content,
-                )
+            )
 
             # Cache expression to avoid duplicate evaluation
             content = nodes.Cache([value], content)
 
             # Define local marker "default"
             content = nodes.Define(
-                [nodes.Alias(["default"], marker("default"))],
+                [nodes.Alias(["default"], self.default_marker)],
                 content
-                )
+            )
 
         return content
 
-    def _create_attributes_nodes(self, prepared, I18N_ATTRIBUTES):
+    def _create_attributes_nodes(self, prepared, I18N_ATTRIBUTES, STATIC):
         attributes = []
 
-        for name, text, quote, space, eq, expr in prepared:
-            implicit_i18n = name.lower() in self.implicit_i18n_attributes
+        names = [attr[0] for attr in prepared]
+        filtering = [[]]
+
+        for i, (name, text, quote, space, eq, expr) in enumerate(prepared):
+            implicit_i18n = (
+                name is not None and
+                name.lower() in self.implicit_i18n_attributes
+            )
 
             char_escape = ('&', '<', '>', quote)
-
-            # Use a provided default text as the default marker
-            # (aliased to the name ``default``), otherwise use the
-            # program's default marker value.
-            if text is not None:
-                default_marker = ast.Str(s=text)
-            else:
-                default_marker = self.default_marker
-
             msgid = I18N_ATTRIBUTES.get(name, missing)
 
             # If (by heuristic) ``text`` contains one or more
             # interpolation expressions, apply interpolation
-            # substitution to the text
+            # substitution to the text.
             if expr is None and text is not None and '${' in text:
-                expr = nodes.Substitution(text, char_escape, None)
+                default = None
+                expr = nodes.Substitution(
+                    text, char_escape, default, self.default_marker)
                 translation = implicit_i18n and msgid is missing
                 value = nodes.Interpolation(expr, True, translation)
-                default_marker = self.default_marker
-
-            # If the expression is non-trivial, the attribute is
-            # dynamic (computed).
-            elif expr is not None:
-                if name in self.boolean_attributes:
-                    value = nodes.Boolean(expr, name)
-                else:
-                    if text is not None:
-                        default = default_marker
-                    else:
-                        default = None
-
-                    value = nodes.Substitution(expr, char_escape, default)
-
-            # Otherwise, it's a static attribute.
             else:
-                value = ast.Str(s=text)
-                if msgid is missing and implicit_i18n:
-                    msgid = text
+                default = ast.Str(s=text) if text is not None else None
 
-            # If translation is required, wrap in a translation
-            # clause
-            if msgid is not missing:
-                value = nodes.Translate(msgid, value)
+                # If the expression is non-trivial, the attribute is
+                # dynamic (computed).
+                if expr is not None:
+                    if name is None:
+                        expression = nodes.Value(
+                            decode_htmlentities(expr),
+                            default,
+                            self.default_marker
+                        )
+                        value = nodes.DictAttributes(
+                            expression, ('&', '<', '>', '"'), '"',
+                            set(filter(None, names[i:]))
+                        )
+                        for fs in filtering:
+                            fs.append(expression)
+                        filtering.append([])
+                    elif name in self.boolean_attributes:
+                        value = nodes.Boolean(
+                            expr, name, default, self.default_marker)
+                    else:
+                        value = nodes.Substitution(
+                            decode_htmlentities(expr),
+                            char_escape,
+                            default,
+                            self.default_marker,
+                        )
 
-            space = self._maybe_trim(space) or " "
-            attribute = nodes.Attribute(name, value, quote, eq, space)
+                # Otherwise, it's a static attribute. We don't include it
+                # here if there's one or more "computed" attributes
+                # (dynamic, from one or more dict values).
+                else:
+                    value = ast.Str(s=text)
+                    if msgid is missing and implicit_i18n:
+                        msgid = text
 
-            # Always define a ``default`` alias
-            attribute = nodes.Define(
-                [nodes.Alias(["default"], default_marker)],
-                attribute,
+            if name is not None:
+                # If translation is required, wrap in a translation
+                # clause
+                if msgid is not missing:
+                    value = nodes.Translate(msgid, value)
+
+                space = self._maybe_trim(space)
+
+                attribute = nodes.Attribute(
+                    name,
+                    value,
+                    quote,
+                    eq,
+                    space,
+                    default,
+                    filtering[-1],
                 )
 
-            attributes.append(attribute)
+                if not isinstance(value, ast.Str):
+                    # Always define a ``default`` alias for non-static
+                    # expressions.
+                    attribute = nodes.Define(
+                        [nodes.Alias(["default"], self.default_marker)],
+                        attribute,
+                    )
 
-        return attributes
+                value = attribute
+
+            attributes.append(value)
+
+        result = nodes.Sequence(attributes)
+
+        # We're caching all expressions found during attribute processing to
+        # enable the filtering functionality which exists to allow later
+        # definitions to override previous ones.
+        expressions = filtering[0]
+        if expressions:
+            return nodes.Cache(expressions, result)
+
+        return result
 
     def _create_static_attributes(self, prepared):
         static_attrs = {}
 
         for name, text, quote, space, eq, expr in prepared:
+            if name is None:
+                continue
+
             static_attrs[name] = text if text is not None else expr
 
+        if not static_attrs:
+            return
+
         return Static(parse(repr(static_attrs)).body)
 
     def _maybe_trim(self, string):
         if self.trim_attribute_space:
-            return string.strip()
+            return re_trim.sub(" ", string)
 
         return string
