@@ -1,33 +1,26 @@
-try:
-    import ast
-except ImportError:
-    from chameleon import ast24 as ast
-
 from functools import partial
+from hashlib import sha256
 from os.path import dirname
 
-from ..i18n import fast_translate
+from ..astutil import Symbol
+from ..compiler import ExpressionEngine
+from ..i18n import simple_translate
+from ..loader import TemplateLoader
+from ..tal import RepeatDict
+from ..tales import DEFAULT_MARKER
+from ..tales import ExistsExpr
+from ..tales import ExpressionParser
+from ..tales import ImportExpr
+from ..tales import NotExpr
+from ..tales import ProxyExpr
 from ..tales import PythonExpr
 from ..tales import StringExpr
-from ..tales import NotExpr
-from ..tales import ExistsExpr
-from ..tales import ImportExpr
-from ..tales import ProxyExpr
 from ..tales import StructureExpr
-from ..tales import ExpressionParser
-
-from ..tal import RepeatDict
-
 from ..template import BaseTemplate
 from ..template import BaseTemplateFile
-from ..compiler import ExpressionEngine
-from ..loader import TemplateLoader
-from ..astutil import Builtin
-from ..utils import decode_string
-from ..utils import string_type
-
 from .program import MacroProgram
 
+
 try:
     bytes
 except NameError:
@@ -43,6 +36,12 @@ class PageTemplate(BaseTemplate):
 
     Configuration (keyword arguments):
 
+      ``auto_reload``
+
+        Enables automatic reload of templates. This is mostly useful
+        in a development mode since it takes a significant performance
+        hit.
+
       ``default_expression``
 
         Set the default expression type. The default setting is
@@ -50,20 +49,11 @@ class PageTemplate(BaseTemplate):
 
       ``encoding``
 
-        The default text substitution value is a unicode string on
-        Python 2 or simply string on Python 3.
+        The default text substitution value is a string.
 
         Pass an encoding to allow encoded byte string input
         (e.g. UTF-8).
 
-      ``literal_false``
-
-        Attributes are not dropped for a value of ``False``. Instead,
-        the value is coerced to a string.
-
-        This setting exists to provide compatibility with the
-        reference implementation.
-
       ``boolean_attributes``
 
         Attributes included in this set are treated as booleans: if a
@@ -89,7 +79,8 @@ class PageTemplate(BaseTemplate):
 
         Example::
 
-          def translate(msgid, domain=None, mapping=None, default=None):
+          def translate(msgid, domain=None, mapping=None, default=None,
+                        context=None):
               ...
               return translation
 
@@ -115,6 +106,11 @@ class PageTemplate(BaseTemplate):
 
           implicit_i18n_attributes = set(['alt', 'title'])
 
+      ``on_error_handler``
+
+        This is an optional exception handler that is invoked during the
+        "on-error" fallback block.
+
       ``strict``
 
         Enabled by default. If disabled, expressions are only required
@@ -128,7 +124,41 @@ class PageTemplate(BaseTemplate):
 
         If set, additional attribute whitespace will be stripped.
 
-    Output is unicode on Python 2 and string on Python 3.
+      ``restricted_namespace``
+
+        True by default. If set False, ignored all namespace except chameleon
+        default namespaces. It will be useful working with attributes based
+        javascript template renderer like VueJS.
+
+        Example:
+
+          <div v-bind:id="dynamicId"></div>
+          <button v-on:click="greet">Greet</button>
+          <button @click="greet">Greet</button>
+
+      ``tokenizer``
+
+        None by default. If provided, this tokenizer is used instead of the
+        default (which is selected based on the template mode parameter.)
+
+      ``value_repr``
+
+        This can be used to override the default value representation
+        function which is used to format values when formatting an
+        exception output. The function must not raise an exception (it
+        should be safe to call with any value).
+
+      ``default_marker``
+
+        This default marker is used as the marker object bound to the `default`
+        name available to any expression. The semantics is such that if an
+        expression evaluates to the marker object, the default action is used;
+        for an attribute expression, this is the static attribute text; for an
+        element this is the static element text. If there is no static text
+        then the default action is similar to an expression result of `None`.
+
+    Output is of type ``str``.
+
     """
 
     expression_types = {
@@ -138,18 +168,16 @@ class PageTemplate(BaseTemplate):
         'exists': ExistsExpr,
         'import': ImportExpr,
         'structure': StructureExpr,
-        }
+    }
 
     default_expression = 'python'
 
-    translate = staticmethod(fast_translate)
+    translate = staticmethod(simple_translate)
 
     encoding = None
 
     boolean_attributes = set()
 
-    literal_false = False
-
     mode = "xml"
 
     implicit_i18n_translate = False
@@ -158,9 +186,21 @@ class PageTemplate(BaseTemplate):
 
     trim_attribute_space = False
 
+    enable_data_attributes = False
+
+    enable_comment_interpolation = True
+
+    on_error_handler = None
+
+    restricted_namespace = True
+
+    tokenizer = None
+
+    default_marker = Symbol(DEFAULT_MARKER)
+
     def __init__(self, body, **config):
         self.macros = Macros(self)
-        super(PageTemplate, self).__init__(body, **config)
+        super().__init__(body, **config)
 
     def __getitem__(self, name):
         return self.macros[name]
@@ -171,102 +211,121 @@ class PageTemplate(BaseTemplate):
 
     @property
     def engine(self):
-        if self.literal_false:
-            default_marker = ast.Str(s="__default__")
-        else:
-            default_marker = Builtin("False")
-
         return partial(
             ExpressionEngine,
             self.expression_parser,
-            default_marker=default_marker,
-            )
+            default_marker=self.default_marker,
+        )
 
     @property
     def expression_parser(self):
         return ExpressionParser(self.expression_types, self.default_expression)
 
     def parse(self, body):
-        if self.literal_false:
-            default_marker = ast.Str(s="__default__")
-        else:
-            default_marker = Builtin("False")
-
         return MacroProgram(
             body, self.mode, self.filename,
             escape=True if self.mode == "xml" else False,
-            default_marker=default_marker,
+            default_marker=self.default_marker,
             boolean_attributes=self.boolean_attributes,
             implicit_i18n_translate=self.implicit_i18n_translate,
             implicit_i18n_attributes=self.implicit_i18n_attributes,
             trim_attribute_space=self.trim_attribute_space,
-            )
+            enable_data_attributes=self.enable_data_attributes,
+            enable_comment_interpolation=self.enable_comment_interpolation,
+            restricted_namespace=self.restricted_namespace,
+            tokenizer=self.tokenizer
+        )
 
-    def render(self, encoding=None, translate=None, target_language=None, **vars):
+    def render(self, encoding=None, **_kw):
         """Render template to string.
 
-        The ``encoding`` and ``translate`` arguments are documented in
-        the template class constructor. If passed to this method, they
-        are used instead of the class defaults.
+        If providd, the ``encoding`` argument overrides the template
+        default value.
 
-        Additional arguments:
+        Additional keyword arguments are passed as template variables.
+
+        In addition, some also have a special meaning:
+
+          ``translate``
+
+            This keyword argument will override the default template
+            translate function.
 
           ``target_language``
 
-            This argument will be partially applied to the translation
-            function.
-
-            An alternative is thus to simply provide a custom
-            translation function which includes this information or
-            relies on a different mechanism.
+            This will be used as the default argument to the translate
+            function if no `i18n:target` value is provided.
 
+            If not provided, the `translate` function will need to
+            negotiate a language based on the provided context.
         """
 
-        non_trivial_translate = translate is not None
-        translate = translate if non_trivial_translate else self.translate or \
-                    type(self).translate
+        translate = _kw.get('translate')
+        if translate is None:
+            translate = self.translate
 
-        # Curry language parameter if non-trivial
-        if target_language is not None:
-            translate = partial(translate, target_language=target_language)
+            # This should not be necessary, but we include it for
+            # backward compatibility.
+            if translate is None:
+                translate = type(self).translate
 
         encoding = encoding if encoding is not None else self.encoding
         if encoding is not None:
-            txl = translate
-
-            def translate(msgid, **kwargs):
+            def translate(msgid, txl=translate, encoding=encoding, **kwargs):
                 if isinstance(msgid, bytes):
-                    msgid = decode_string(msgid, encoding)
+                    msgid = bytes.decode(msgid, encoding)
                 return txl(msgid, **kwargs)
 
-            def decode(inst):
-                return decode_string(inst, encoding, 'ignore')
+            def decode(inst, encoding=encoding):
+                return bytes.decode(inst, encoding, 'ignore')
         else:
-            decode = decode_string
+            decode = bytes.decode
 
-        setdefault = vars.setdefault
+        target_language = _kw.get('target_language')
+
+        setdefault = _kw.setdefault
         setdefault("__translate", translate)
-        setdefault("__convert", translate)
+        setdefault("__convert",
+                   partial(translate, target_language=target_language))
         setdefault("__decode", decode)
-
-        if non_trivial_translate:
-            vars['translate'] = translate
+        setdefault("target_language", None)
+        setdefault("__on_error_handler", self.on_error_handler)
 
         # Make sure we have a repeat dictionary
-        if 'repeat' not in vars: vars['repeat'] = RepeatDict({})
+        if 'repeat' not in _kw:
+            _kw['repeat'] = RepeatDict({})
 
-        return super(PageTemplate, self).render(**vars)
+        return super().render(**_kw)
 
     def include(self, *args, **kwargs):
         self.cook_check()
         self._render(*args, **kwargs)
 
+    def digest(self, body, names):
+        hex = super().digest(body, names)
+        if isinstance(hex, str):
+            hex = hex.encode('utf-8')
+        digest = sha256(hex)
+        digest.update(';'.join(names).encode('utf-8'))
+
+        for attr in (
+            'trim_attribute_space',
+            'implicit_i18n_translate',
+            'strict'
+        ):
+            v = getattr(self, attr)
+            digest.update(
+                (";{}={}".format(attr, str(v))).encode('ascii')
+            )
+
+            return digest.hexdigest()[:32]
+
     def _builtins(self):
         return {
             'template': self,
             'macros': self.macros,
             'nothing': None,
-            }
+        }
 
 
 class PageTemplateFile(PageTemplate, BaseTemplateFile):
@@ -282,7 +341,7 @@ class PageTemplateFile(PageTemplate, BaseTemplateFile):
 
     Below are listed the configuration arguments specific to
     file-based templates; see the string-based template class for
-    general options and documentation::
+    general options and documentation:
 
     Configuration (keyword arguments):
 
@@ -313,37 +372,45 @@ class PageTemplateFile(PageTemplate, BaseTemplateFile):
     """
 
     expression_types = PageTemplate.expression_types.copy()
-    expression_types['load'] = partial(ProxyExpr, '__loader')
+    expression_types['load'] = partial(
+        ProxyExpr, '__loader',
+        ignore_prefix=False
+    )
 
     prepend_relative_search_path = True
 
     def __init__(self, filename, search_path=None, loader_class=TemplateLoader,
                  **config):
-        super(PageTemplateFile, self).__init__(filename, **config)
-
         if search_path is None:
             search_path = []
         else:
-            if isinstance(search_path, string_type):
+            if isinstance(search_path, str):
                 search_path = [search_path]
             else:
                 search_path = list(search_path)
 
-        # If the flag is set (this is the default), prepend the path
-        # relative to the template file to the search path
-        if self.prepend_relative_search_path:
-            path = dirname(self.filename)
-            search_path.insert(0, path)
+        def post_init():
+            # If the flag is set (this is the default), prepend the path
+            # relative to the template file to the search path
+            if self.prepend_relative_search_path:
+                path = dirname(self.filename)
+                search_path.insert(0, path)
 
-        loader = loader_class(search_path=search_path, **config)
-        template_class = type(self)
+            loader = loader_class(search_path=search_path, **config)
+            template_class = type(self)
 
-        # Bind relative template loader instance to the same template
-        # class, providing the same keyword arguments.
-        self._loader = loader.bind(template_class)
+            # Bind relative template loader instance to the same template
+            # class, providing the same keyword arguments.
+            self._loader = loader.bind(template_class)
+
+        super().__init__(
+            filename,
+            post_init_hook=post_init,
+            **config
+        )
 
     def _builtins(self):
-        d = super(PageTemplateFile, self)._builtins()
+        d = super()._builtins()
         d['__loader'] = self._loader
         return d
 
@@ -368,18 +435,18 @@ class PageTextTemplateFile(PageTemplateFile):
     mode = "text"
 
     def render(self, **vars):
-        result = super(PageTextTemplateFile, self).render(**vars)
+        result = super().render(**vars)
         return result.encode(self.encoding or 'utf-8')
 
 
-class Macro(object):
+class Macro:
     __slots__ = "include",
 
     def __init__(self, render):
         self.include = render
 
 
-class Macros(object):
+class Macros:
     __slots__ = "template",
 
     def __init__(self, template):
